;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

;; uservar order is IMPORTANT!

" i-next" z80-labman:@get constant NEXT
" i-next-push-hl" z80-labman:@get constant PUSHHL
" i-next-push-de-hl" z80-labman:@get constant PUSHDE

zx-s0 tcom:zx-uservar (S0)
zx-r0 tcom:zx-uservar (R0)

\ 0 uservar FLD -- not used in FIG
10 uservar BASE
 0 uservar BLK
 0 uservar DPL
 0 uservar HLD

0 uservar IN
zx-query-tib uservar TIB

;; this is not used by the system.
;; intended to hold a pointer to task data.
0 uservar TSK

;; i'm not sure about user/globals yet...
0 variable R# -- this is only used in EDITOR
0 variable SCR  -- used by LIST and EDITOR

;; no more uservars beyond this point


tcom:zx-dp^ constant (DP)
tcom:zx-state^ constant STATE
tcom:zx-context^ constant CONTEXT
tcom:zx-current^ constant CURRENT
tcom:zx-fence^ constant FENCE
tcom:zx-voclink^ constant VOC-LINK

zx-first-buffer constant FIRST
zx-buffer-limit constant LIMIT

zx-buffer-lo constant LO
zx-buffer-hi constant HI

0 variable OFFSET
;; will be reset in COLD
zx-first-buffer variable PREV
zx-first-buffer variable USE

0 constant 0
1 constant 1
2 constant 2
3 constant 3

32 constant BL
10 constant NL
64 constant C/L

128 constant B/BUF
8 constant B/SCR
8 constant #BUFF

31 variable WIDTH
0 variable WARNING

0 variable CSP


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lit, branch

code: LIT  ( -- n )
  ld    a, (bc)
  inc   bc
  ld    l, a
  ld    a, (bc)
  inc   bc
  ld    h, a
;code-push-hl


OPT-CLIT? [IF]
code: CLIT  ( -- n )
  ld    h, # 0
zx-word-clit-finish:
  ld    a, (bc)
  inc   bc
  ld    l, a
;code-push-hl

code: BLIT  ( -- n )
  ld    h, # $FF
  jr    # zx-word-clit-finish
;code-no-next
[ENDIF]


code: BRANCH
  ld    hl, bc
 OPT-REL-BRANCH? [IF]
  ld    e, (hl)
  inc   hl
  ld    d, (hl)
  dec   hl
  add   hl, de
  ld    bc, hl
 [ELSE]
  ld    c, (hl)
  inc   hl
  ld    b, (hl)
 [ENDIF]
;code

code: 0BRANCH  ( n )
  pop   hl
  ld    a, l
  or    h
  jr    z, # zx-word-branch
  inc   bc
  inc   bc
;code

code: TBRANCH  ( n )
  pop   hl
  ld    a, l
  or    h
  jr    nz, # zx-word-branch
  inc   bc
  inc   bc
;code


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; exit

code: ;S
zx-word-s-semi:
  ld    hl, () zx-rp
  ld    c, (hl)
  inc   hl
  ld    b, (hl)
  inc   hl
  ld    zx-rp (), hl
;code

code: EXIT
  jr    # zx-word-s-semi
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic stack manipulations

code: SP@  ( -- sp )
  ld    hl, # 0
  add   hl, sp
;code-push-hl

code: SP!
  ld    l, (ix+) zx-uvar-ofs (S0)
  ld    h, (ix+) zx-uvar-ofs (S0) 1+
  ld    sp, hl
;code

code: DEPTH  ( -- stack-depth )
  ld    l, (ix+) zx-uvar-ofs (S0)
  ld    h, (ix+) zx-uvar-ofs (S0) 1+
  or    a
  sbc   hl, sp
  rr    h
  rr    l
;code-push-hl

code: RP@  ( -- rp )
  ld    hl, () zx-rp
;code-push-hl

code: RP!
  ld    l, (ix+) zx-uvar-ofs (R0)
  ld    h, (ix+) zx-uvar-ofs (R0) 1+
  ld    zx-rp (), hl
;code


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic stack manipulations

code: 2DUP  ( a b -- a b a b )
  pop   hl  ;; b
  pop   de  ;; a
  push  de
  push  hl
;code-push-de-hl

code: 2DROP  ( a b )
  pop   hl
  pop   hl
;code

\ code: 4DROP  ( a b c d )
\   ;; almost 2 times faster than 4 POPs
\   ld    hl, # 8   ;; 10
\   add   hl, sp    ;; 11
\   ld    sp, hl    ;; 6
\   ;; 27
\ ;code

;; k8: sorry!
code: 2SWAP  ( a b c d -- c d a b )
  exx
  pop   de  ;; d
  pop   bc  ;; c
  exx
  pop   hl  ;; b
  pop   de  ;; a
zx-word-2swap-finish:
  exx
  push  bc
  push  de
  exx
;code-push-de-hl

;; k8: sorry!
code: 2OVER  ( a b c d -- a b c d a b )
  exx
  pop   de  ;; d
  pop   bc  ;; c
  exx
  pop   hl  ;; b
  pop   de  ;; a
  push  hl
  push  de
  jr    # zx-word-2swap-finish
;code-no-next

;; SWAP DROP
code: NIP  ( a b -- b )
  pop   hl
  pop   de
;code-push-hl

;; SWAP OVER
code: TUCK  ( a b -- b a b )
  pop   hl  ;; b
  pop   de  ;; a
  push  hl
;code-push-de-hl

code: -DUP  ( a -- a a // 0 )
  pop   hl
  push  hl
  ld    a, h
  or    l
  jp    z, # i-next
;code-push-hl

code: -ROT  ( a b c -- c a b )
  pop   af  ;; c
  pop   hl  ;; b
  pop   de  ;; a
  jr    # zx-word-rot-finish
;code-no-next

code: ROT  ( a b c -- b c a )
  pop   de  ;; c
  pop   af  ;; b
  pop   hl  ;; a
zx-word-rot-finish:
  push  af
;code-push-de-hl

code: OVER  ( a b -- a b a )
  pop   de
  pop   hl
  push  hl
;code-push-de-hl

code: SWAP  ( a b -- b a )
  pop   hl
  ex    (sp), hl
;code-push-hl

code: DROP  ( a )
  pop   hl
;code

code: DUP  ( a -- a a )
  pop   hl
  push  hl
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; execute

code: EXECUTE  ( cfa )
  pop   hl
 OPT-DTC? [IF]
  jp    hl
 [ELSE]
  jp    # i-next-cfa-hl
 [ENDIF]
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; loops

code: (LOOP)
  ld    de, # 1
zx-word-addloop-do:
  ld    hl, () zx-rp
  ld    a, (hl)
  add   a, e
  ld    (hl), a
  ld    e, a
  inc   hl
  ld    a, (hl)
  adc   a, d
  ld    (hl), a
  inc   hl
  inc   d
  dec   d
  ld    d, a
  jp    m, # .skip0
  ld    a, e
  sub   (hl)
  ld    a, d
  inc   hl
  sbc   a, (hl)
  jp    # .skip1
.skip0:
  ld    a, (hl)
  sub   e
  inc   hl
  ld    a, (hl)
  sbc   a, d
.skip1:
  jp    m, # zx-word-branch
  inc   hl
  ld    zx-rp (), hl
  inc   bc
  inc   bc
;code

code: (+LOOP)  ( inc )
  pop   de
  jr    # zx-word-addloop-do
;code-no-next

code: (DO)  ( limit start )
  ld    hl, () zx-rp
  ld    de, # -4
  add   hl, de
  ld    zx-rp (), hl
  pop   de
  ld    (hl), e
  inc   hl
  ld    (hl), d
  pop   de
  inc   hl
  ld    (hl), e
  inc   hl
  ld    (hl), d
;code

code: LEAVE
  ld    hl, () zx-rp
  ld    e, (hl)
  inc   hl
  ld    d, (hl)
  inc   hl
  ld    (hl), e
  inc   hl
  ld    (hl), d
;code

code: I'  ( -- n )
  ld    hl, () zx-rp
  inc   hl
  inc   hl
zx-load-hl-word-and-push:
  ld    e, (hl)
  inc   hl
  ld    d, (hl)
  push  de
;code

code: J  ( -- n )
  ld    hl, () zx-rp
  ld    de, # 4
  add   hl, de
  jr    # zx-load-hl-word-and-push
;code-no-next

code: I  ( -- loop-counter )
  ld    hl, () zx-rp
  jr    # zx-load-hl-word-and-push
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; return stack push, pop, peek, drop

code: R  ( | n -- n | n )
  ld    hl, () zx-rp
  jr    # zx-load-hl-word-and-push
;code-no-next

code: R@  ( | n -- n | n )
  ld    hl, () zx-rp
  jr    # zx-load-hl-word-and-push
;code-no-next

code: >R  ( n -- | n )
  pop   de
  ld    hl, () zx-rp
  dec   hl
  ld    (hl), d
  dec   hl
  ld    (hl), e
zx-next-store-hl-to-rp:
  ld    zx-rp (), hl
;code

code: R>  ( | n -- n )
  ld    hl, () zx-rp
  ld    e, (hl)
  inc   hl
  ld    d, (hl)
  inc   hl
  push de
  jr    # zx-next-store-hl-to-rp
;code-no-next

code: RDROP  ( | n -- )
  ld    hl, () zx-rp
  inc   hl
  inc   hl
  jr    # zx-next-store-hl-to-rp
;code-no-next



;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; digit conversion

code: DIGIT  ( c n1 -- n2 TRUE // FALSE )
  pop   hl
  pop   de
  ld    a, e
  sub   a, # $30
  jp    m, # .done
  cp    # 10
  jp    m, # .skip
  sub   a, # 7
  cp    # 10
  jp    m, # .done
.skip:
  cp    l
  jp    p, # .done
  ld    e, a
  ld    hl, # 1
  next-push-de-hl
.done:
  ld    l, h
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; dictionary searching

;; Searches the dictionary starting at the name field address addr2, matching
;; to the text at addr1. Returns parameter field address, length byte of name
;; field and true for a good match. If no match is found, only a boolean false
;; is left.
code: (FIND)  ( addr1 addr2 -- pfa b TRUE // FALSE )
  pop   de
.main-loop:
  pop   hl
  push  hl
  ;; DE=vocab-nfa
  ;; HL=name-cstr
  ld    a, (de)
  xor   (hl)
  and   # $3F
  jr    nz, # .skip-nfa
.name-cmp-loop:
  inc   hl
  inc   de
  ld    a, (de)
  xor   (hl)
  add   a, a
  jr    nz, # .wrong-name
  jr    nc, # .name-cmp-loop
  ;; HL is at the end of NFA
 OPT-DTC? [IF]
  ld    hl, # 6
 [ELSE]
  ld    hl, # 5
 [ENDIF]
  add   hl, de
  ex    (sp), hl
.nfa-back:
  dec   de
  ld    a, (de)
  or    a
  jp    p, # .nfa-back
  ld    e, a
  ld    d, # 0
  ld    hl, # 1
  next-push-de-hl
.wrong-name:
  jr    c, # .at-lfa
.skip-nfa:
  inc   de
  ld    a, (de)
  or    a
  jp    p, # .skip-nfa
.at-lfa:
  inc   de
  ex    de, hl
  ld    e, (hl)
  inc   hl
  ld    d, (hl)
  ld    a, d
  or    e
  jr    nz, # .main-loop
  pop   hl
  ld    hl, # 0
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; input buffer scanning

;; The text scanning primitive used by WORD. From the text address addr1 and
;; an ascii delimiting character c, is determined the byte offset to the first
;; non-delimiter character n1, the offset to the first delimiter after the
;; text n2, and the offset to the first character not included. This procedure
;; will not process past an ascii "null", treating it as an unconditional
;; delimiter.
code: ENCLOSE  ( addr1 c -- addr1 n1 n2 n3 )
  pop   de
  pop   hl
  push  hl
  ld    a, e
  ld    d, a
  ld    e, # $FF
  dec   hl
.l5fc6h:
  inc   hl
  inc   e
  cp    (hl)
  jr    z, # .l5fc6h
  ld    d, # 0
  push  de
  ld    d, a
  ld    a, (hl)
  and   a
  jr    nz, # .l5fdch
  ld    d, # 0
  inc   e
  push  de
  dec   e
  \ push  de
  \ next
  jr    # .done-push-de
.l5fdch:
  ld    a, d
  inc   hl
  inc   e
  cp    (hl)
  jr    z, # .l5fedh
  ld    a, (hl)
  and   a
  jr    nz, # .l5fdch
  ld    d, # 0
  push  de
  \ push  de
  \ next
  jr    # .done-push-de
.l5fedh:
  ld    d, # 0
  push  de
  inc   e
.done-push-de:
  push  de
;code


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; memory copying and filling

code: CMOVE  ( asrc adest acount )
  ld    hl, bc
  pop   bc        ;; count
  pop   de        ;; dest
  ex    (sp), hl  ;; src
  ld    a, b
  or    c
  jr    z, # .skip
  ldir
.skip:
  pop   bc
;code

code: FILL  ( addr len byte )
  ld    hl, bc
  pop   de
  pop   bc
  ex    (sp), hl
  ;; HL: addr
  ;; DE: byte
  ;; BC: len
  ld    a, b
  or    a
  jr    nz, # .doit
  dec   c
  jp    m, # .done
  jr    z, # .one-byte
  inc   c
.doit:
  ld    (hl), e
  dec   bc
  ld    de, hl
  inc   de
  ldir
.done:
  pop   bc
  next
.one-byte:
  ld    (hl), e
  jr    # .done
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; debug primitives

OPT-DOT-HEX? [IF]
code: .HEX2  ( n )
  pop   hl
zx-do-hex2:
  ld    a, l
  rrca
  rrca
  rrca
  rrca
  call  # zx-emit-hex-digit
  ld    a, l
  call  # zx-emit-hex-digit
  next
zx-emit-hex-digit:
  push  bc
  push  hl
  push  de
  and   # $0F
  add   a, # [char] 0
  cp    # [char] 9 1+
  jr    c, # .skip
  add   a, # [char] A [char] 9 1+ -
.skip:
  rst   # $10
  pop   de
  pop   hl
  pop   bc
  ret
;code-no-next

code: .HEX4  ( n )
  pop   hl
  ld    a, h
  rrca
  rrca
  rrca
  rrca
  call  # zx-emit-hex-digit
  ld    a, h
  call  # zx-emit-hex-digit
  jr    # zx-do-hex2
;code-no-next
[ENDIF]


<zx-done>
