;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; input/output primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

0 variable OBL? -- last printed char was BL?
0 variable OUT

(*
<zx-hidden>
code: OPEN-CHAN-#2
  push  bc
  ld    a, # 2
  call  # $1601
  pop   bc
;code
<zx-normal>
*)


;; original printer: $094F
;; waiting for one operand printer: $0A87
;; waiting for two operands printer: $0A6D
;; print routine is at 23633


;; k8: removed printer output
;; k8: also, removed channel open on each emit
code: (EMIT)  ( ch )
  ;; load printing routine high address byte
  ld    hl, () 23633
  inc   hl
  ld    a, (hl)
  ;; load arguments
  pop   hl
  push  bc
  ;; check if we are printing something special
  cp    # $09
  ld    a, l
  jr    nz, # .print-special
  ;; 0x02 is "ENDCR"
  cp    # $02
  jr    z, # .do-endcr
  ;; blank check
  ld    hl, # zx-['pfa] OBL?
  ld    (hl), # 0
  cp    # 32
  jr    nz, # .not-bl
  inc   (hl)
.not-bl:
  ;; increment "OUT"?
  cp    # 32
  jr    c, # .no-out-inc
  ld    hl, # zx-['pfa] OUT
  inc   (hl)
.no-out-inc:
  cp    # 10
  jr    nz, # .do-print
  ld    a, # 13
.do-print:
  cp    # 7
  jr    nz, # .not-edit
  ld    a, # [char] ?
.not-edit:
.print-special:
  rst   # $10
  \ call  # $09F4
  ;; reset scroll
  ld    a, # $FF
  ld    $5C8C (), a
  jr    # .exit
.do-endcr:
  ld    a, () 23688
  sub   # 33
  ld    a, # 13
  jr    nz, # .print-special
.exit:
  pop   bc
;code


;; doesn't wait
code: (KEY-NW)  ( -- char // 0 )
  \ push  bc
  \ ld    a, # 2
  \ call  # $1601
  \ pop   bc
  ei    ;; just in case

  ;; CUR_MODE
  ld    7 (iy+), # 00 ;; #5C41

  halt
  ;;;call #02BF       ;; keyboard
  ld    hl, # $5C08   ;; last_k
  ld    a, (hl)
  ld    (hl), # 0
  or    a
  ;; carry is not set, so A value doesn't matter
  jr    z, # .done-zero

  ;; something's pressed
  cp    # 6           ;; CS+2?
  jr    nz, # .not-caps
  ;; toggle CAPS
  ;;ld    hl, # $5C6A   ;; flags
  ld    l, # $6A
  ld    a, # $08      ;; CAPS flag
  xor   (hl)
  ld    (hl), a
  ;; carry is not set, so A value doesn't matter
  jr    # .done-zero

.not-caps:
  cp    # 128
  jr    c, # .done-zero

  ;; translate codes
  ld    e, a
  ld    hl, # key-translation-table 1-
.table-loop:
  inc   hl
  ld    a, (hl)       ;; src
  or    a             ;; end?
  jr    z, # .table-not-found
  inc   hl
  cp    e
  ld    a, (hl)       ;; dest
  jr    nz, # .table-loop
  ld    e, a

  ;; check codes
.table-not-found:
  ld    a, e
.no-trans:
  cp    # 128
.done-zero:
  ld    hl, # 0
  ;; we may jump here from last_k check; carry is reset in this case
  jr    nc, # .done
  ld    l, a
  ;; ENTER produces #13
  ;; SS+ENTER produces #2
  cp    # 13
  jr    nz, # .done
  ;; check for SS pressed
  ld    a, # $7F
  in    a, () $FE
  and   # $02
  jr    nz, # .done
  ld    l, # 2

.done:
  next-push-hl
flush!

key-translation-table:
  198 db,  91 db, ;; AND->[
  197 db,  93 db, ;; OR->]
  172 db, 127 db, ;; AT->(c)
  226 db, 126 db, ;; STOP->~
  195 db, 124 db, ;; NOT->|
  205 db,  92 db, ;; STEP->slash
  204 db, 123 db, ;; TO->{
  203 db, 125 db, ;; THEN->}
0 db,
;code-no-next


;; draw input cursor
<zx-hidden>
code: (.CUR)
  push  bc
  ld    a, () $5C6A
  and   # $08
  ld    a, # [char] L
  jr    z, # .oklow
  ld    a, # [char] C
.oklow:
  call  # $18C1   ;; out-flash
zx-word-cur-bs:
  ld    a, # 8
  \ rst   # $10
  call  # $09F4
  pop   bc
;code

;; erase input cursor
code: (.0CUR)
  push  bc
  ld    a, # 32
  \ rst   # $10
  call  # $09F4
  jr    # zx-word-cur-bs
;code-no-next
<zx-normal>


code: ?TERMINAL  ( -- break? )
  call  # $1F54   ;; this doesn't modify BC
  ld    a, # 0
  ld    h, a
  ccf
  adc   a, h
  ld    l, a
;code-push-hl


<zx-done>
