;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; various math and compare primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; bitwise logic

code: AND  ( a b -- a&b )
  pop   de
  pop   hl
  ld    a, e
  and   l
  ld    l, a
  ld    a, d
  and   h
  ld    h, a
;code-push-hl

code: OR  ( a b -- a|b )
  pop   de
  pop   hl
  ld    a, e
  or    l
  ld    l, a
  ld    a, d
  or    h
  ld    h, a
;code-push-hl

code: XOR  ( a b -- a|b )
  pop   de
  pop   hl
  ld    a, e
  xor   l
  ld    l, a
  ld    a, d
  xor   h
  ld    h, a
;code-push-hl


code: LO-BYTE  ( u -- u>>8 )
  pop   hl
zx-word-lo-byte-finish:
  ld    h, # 0
;code-push-hl

code: HI-BYTE  ( u -- u>>8 )
  pop   hl
  ld    l, h
  jr    # zx-word-lo-byte-finish
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; add, sub

code: +  ( a b -- a+b )
  pop   de
  pop   hl
  add   hl, de
;code-push-hl

code: -  ( a b -- a-b )
  pop   de
  pop   hl
  or    a
  sbc   hl, de
;code-push-hl


code: D+  ( da db -- da+db )
  ld    hl, # 6
  add   hl, sp
  ld    e, (hl)
  ld    (hl), c
  inc   hl
  ld    d, (hl)
  ld    (hl), b
  pop   bc
  pop   hl
  add   hl, de
  ex    de, hl
  pop   hl
  adc   hl, bc
  pop   bc
;code-push-de-hl


code: 1+  ( a -- a+1)
  pop   hl
zx-word-inc-finish:
  inc   hl
;code-push-hl

code: 2+  ( a -- a+2)
  pop   hl
  inc   hl
  jr    # zx-word-inc-finish
;code-no-next

code: 1-  ( a -- a-1)
  pop   hl
zx-word-dec-finish:
  dec   hl
;code-push-hl

code: 2-  ( a -- a-2)
  pop   hl
  dec   hl
  jr    # zx-word-dec-finish
;code-no-next


code: 2*  ( n -- n*2 )
  pop   hl
  add   hl, hl
;code-push-hl

zx-code: 2/  ( n -- n/2 )
  pop   hl
  ld    a, h
  cp    # 128
  ccf
:two-div-rest
  rr    h
  rr    l
;code-push-hl

zx-code: 2u/  ( n -- n/2 )
  pop   hl
  or    a
  jr    # two-div-rest
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; negate

code: NEGATE  ( a -- negated-a )
  pop   de
  ld    hl, # 0
  or    a
  sbc   hl, de
;code-push-hl

code: MINUS  ( a -- negated-a )
  jr    # zx-word-negate
;code-no-next


code: DNEGATE  ( da -- negated-da )
  pop   hl
  pop   de
  sub   a
  sub   e
  ld    e, a
  ld    a, # 0
  sbc   a, d
  ld    d, a
  ld    a, # 0
  sbc   a, l
  ld    l, a
  ld    a, # 0
  sbc   a, h
  ld    h, a
;code-push-de-hl

code: DMINUS  ( a -- negated-a )
  jr    # zx-word-dnegate
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; C>S, S->D

code: C>S  ( signed-byte -- n )
  pop   hl
  ld    a, l
  cp    # $80
  ld    a, # $FE
  adc   a, # 1
;code-push-hl

code: S->D  ( n -- d )
  pop   de
  ld    hl, # 0
  ld    a, d
  and   # $80
  jr    z, # .l6e1dh
  dec   hl
.l6e1dh:
;code-push-de-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; comparisons with 0

code: 0=  ( a -- a=0? )
zx-word-0-equ:
  pop   hl
  ld    a, l
  or    h
@zx-next-inc-l-if-z-push-hl:
  ld    hl, # 0
  jr    nz, # .skip
@zx-next-inc-l-push-hl:
  inc   l
.skip:
;code-push-hl

code: 0<>  ( a -- a<>0? )
zx-word-0-notequ:
  pop   hl
  ld    a, l
  or    h
@zx-next-inc-l-if-nz-push-hl:
  ld    hl, # 0
  jr    z, # .skip
  inc   l
.skip:
;code-push-hl

code: NOT  ( a -- a=0? )
  jr    # zx-word-0-equ
;code-no-next

code: 0<  ( a -- a<0? )
  pop   hl
  bit   7, h
  jr    # zx-next-inc-l-if-nz-push-hl
;code-no-next

code: 0>=  ( a -- a>=0? )
  pop   hl
  bit   7, h
  jr    # zx-next-inc-l-if-z-push-hl
;code-no-next

code: 0<=  ( a -- a<=0? )
  pop   hl
  bit   7, h
  jr    nz, # zx-next-inc-l-if-nz-push-hl
  ld    a, h
  or    l
  jr    # zx-next-inc-l-if-z-push-hl
;code-no-next

code: 0>  ( a -- a>0? )
  pop   hl
  jr    z, # zx-next-inc-l-if-nz-push-hl
  ld    a, h
  or    l
  jr    # zx-next-inc-l-if-nz-push-hl
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; numeric comparisons

code: =  ( a b -- a=b? )
  pop   hl
  pop   de
  or    a
  sbc   hl, de
  ld    hl, # 0
  jr    nz, # .skip
  inc   l
.skip:
;code-push-hl

code: <>  ( a b -- a<>b? )
  pop   hl
  pop   de
  or    a
  sbc   hl, de
  ld    hl, # 0
  jr    z, # .skip
  inc   l
.skip:
;code-push-hl

code: <  ( a b -- a<b? )
  pop   de
  pop   hl
zx-word-do-less:
  ld    a, d
  xor   h
  jp    m, # .l6559h
  and   a
  sbc   hl, de
.l6559h:
  inc   h
  dec   h
  ld    hl, # 1
  jp    m, # .skip
  dec   l
.skip:
;code-push-hl

code: >  ( a b -- a>b? )
  pop   hl
  pop   de
  jr    # zx-word-do-less
;code-no-next

code: U<  ( ua ub -- ua<ub? )
  pop   de
  pop   hl
  or    a
  sbc   hl, de
  ld    h, # 0
  ld    a, h
  adc   a, h
  ld    l, a
;code-push-hl


<zx-done>
