;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; high-level low-level words ;-)
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

: NOOP  ;

: ?COMP  STATE @ 0= 17 ?ERROR ;
: ?EXEC  STATE @ 18 ?ERROR ;
: ?PAIRS ( n1 n2 )  - 19 ?ERROR ;
: ?CSP   SP@ CSP @ - 20 ?ERROR ;

: ?LOADING  BLK @ 0= 22 ?ERROR ;

: SMUDGE  LATEST 32 TOGGLE ;
: IMMEDIATE  LATEST 64 TOGGLE ;

: HEX      16 BASE ! ;
: DECIMAL  10 BASE ! ;

: HERE   ( n )  (DP) @ ;
: ALLOT  ( n )  (DP) +! ;

: ,  ( n )  HERE ! 2 ALLOT ;
: C, ( n )  HERE C! 1 ALLOT ;

: CALL,  ( cfa )
  [ OPT-DTC? ] [IF]
    ;; follow JP chain. this makes primitive calls faster.
    BEGIN DUP C@ $C3 - 0WHILE 1+ @ REPEAT
  [ENDIF] , ;

OPT-DTC? [IF]
: LFA  ( pfa -- lfa )  5 - ;
: CFA  ( pfa -- cfa )  3 - ;
: NFA  ( pfa -- nfa )  6 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 6 + ;
[ELSE]
: LFA  ( pfa -- lfa )  4 - ;
: CFA  ( pfa -- cfa )  2- ;
: NFA  ( pfa -- nfa )  5 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 5 + ;
[ENDIF]


;; added by k8
*: RECURSE  LATEST PFA CFA CALL, ;

;; added by k8
*: RECURSE-TAIL  COMPILE BRANCH LATEST PFA BACK ;


;; k8: wow, it is not immediate!
: COMPILE  ?COMP R> DUP 2+ >R @ , ;

*: [COMPILE]  -FIND 0= 0 ?ERROR DROP CFA CALL, ;

OPT-CLIT? [IF]
*: LITERAL
  STATE @ 0IF EXIT ENDIF
  DUP HI-BYTE DUP 0IF DROP COMPILE CLIT C, EXIT ENDIF
  255 - 0IF COMPILE BLIT C, EXIT ENDIF
  COMPILE LIT , ;
[ELSE]
*: LITERAL  STATE @ IF COMPILE LIT , ENDIF ;
[ENDIF]

*: DLITERAL   STATE @ IF SWAP [COMPILE] LITERAL [COMPILE] LITERAL ENDIF ;


: (ABORT)  ABORT ;

: EMIT   ( ch )  (EMIT) ;
: CR     NL EMIT ;
\ : ENDCR  23688 C@ 33 - IF CR ENDIF ;
: ENDCR  2 EMIT ;
: SPACE  BL EMIT ;
: ?SPACE OBL? C@ 0IF SPACE ENDIF ;

<zx-hidden>
code: KEY-BEEP
  push  bc
  push  ix
  ld    hl, # $C8   ;; pitch
  ld    a, () 23609
zx-word-kbeep-finish:
  ld    e, a
  ld    d, # 0
zx-word-kbeep-do-beep:
  call  # $03B5
  pop   ix
  pop   bc
;code

code: ERR-BEEP
  push  bc
  push  ix
  ld    hl, # $1A90 ;; pitch
  ld    a, () 23608
  jr    # zx-word-kbeep-finish
;code-no-next
<zx-normal>


: KEY  ( -- code )
  0 $5C08 C!
  $FF
  BEGIN
    $5C6A C@ 8 AND TUCK - IF (.CUR) ENDIF
    (KEY-NW)
  -DUP UNTIL KEY-BEEP
  NIP (.0CUR) ;


\ : TYPE  ( addr count )
\   -DUP IF OVER + SWAP DO I C@ EMIT LOOP ELSE DROP ENDIF ;
: TYPE  ( addr count )
  \ DUP IF SWAP COUNT EMIT SWAP 1- RECURSE-TAIL ENDIF 2DROP ;
  BEGIN DUP WHILE SWAP C@++ EMIT SWAP 1- REPEAT 2DROP ;

: -TRAILING  ( addr count -- addr1 count1 )
  DUP 0 DO ( OVER OVER)2DUP + ( 1 -)1- C@ BL - IF LEAVE ELSE ( 1 -)1- ENDIF LOOP ;

: (.")  R COUNT DUP 1+ R> + >R TYPE ;

: ERASE  ( addr count )  0 fill ;
: BLANKS ( addr count )  bl fill ;

: PAD  ( -- pad^ )  here 68 + ;


: +-  ( n1 n2 -- n3 )  0< IF MINUS ENDIF ;
: D+- ( d1 n -- d2 )   0< IF DMINUS ENDIF ;

: ABS  ( n1 -- n2 )  DUP +- ;
: DABS ( d1 -- d2 )  DUP D+- ;

: MIN  ( a b -- c )  2DUP > IF SWAP ENDIF DROP ;
: MAX  ( a b -- c )  2DUP < IF SWAP ENDIF DROP ;

: M*  ( nl n2 -- d )  2DUP XOR >R ABS SWAP ABS U* R> D+- ;
: M/  ( d n1 -- n2 n3 )  OVER >R >R DABS R ABS U/MOD R> R XOR +- SWAP R> +- SWAP ;

: *  ( a b -- c )  M* DROP ;
: /MOD  ( a b -- a%b a/b )  >R S->D R> M/ ;
: /  ( a b -- c )  /MOD SWAP DROP ;
: MOD  ( a b -- c )  /MOD DROP ;
: */MOD  ( a b -- a%b a/b )  >R M* R> M/ ;
: */  ( a b -- c )  */MOD SWAP DROP ;
: M/MOD  ( ud1 u2 -- u3 ud4 )  >R 0 R U/MOD R> SWAP >R U/MOD R> ;

: TRAVERSE  ( addr1 n -- addr2 )  SWAP BEGIN OVER + 127 OVER C@ < UNTIL SWAP DROP ;

: LATEST  ( -- nfa )  CURRENT @ @ ;


: +BUF  ( addr1 -- addr2 0-if-prev )
  132 + DUP LIMIT = IF DROP FIRST ENDIF DUP PREV @ - ;

: UPDATE  PREV @ @ -32768 OR PREV @ ! ;

: EMPTY-BUFFERS
  FIRST LIMIT OVER - ERASE LIMIT FIRST DO 32767 I ! 132 +LOOP ;

: DR0  OFFSET OFF ;

: R/W  ( addr blk read-flag )
  >R B/BUF * LO + DUP HI > 6 ?ERROR R> IF SWAP ENDIF B/BUF CMOVE ;

: BUFFER  ( n -- addr )
  USE @ DUP >R BEGIN +BUF UNTIL USE !
  R @ 0< IF R 2+ R @ 32767 AND 0 R/W ENDIF
  R ! R PREV ! R> 2+ ;

: BLOCK  ( n -- addr )
  OFFSET @ + >R PREV @ DUP @ R - DUP + IF
    BEGIN
      +BUF 0= IF DROP R BUFFER DUP R 1 R/W ( 2 -)2- ENDIF
      DUP @ R - DUP + 0=
    UNTIL
    DUP PREV !
  ENDIF
  R> DROP 2+ ;

: FLUSH  #BUFF 1+ 0 DO 0 BUFFER DROP LOOP ;

: ERROR  ( n -- in blk )
  WARNING @ 0< IF (ABORT) ENDIF
  HERE COUNT TYPE ." ? " MESSAGE SP!
  BLK @ -DUP IF IN @ SWAP ENDIF
  QUIT ;

: ?ERROR  ( flag err# )  SWAP IF ERROR ELSE DROP ENDIF ;

: (LINE)  ( n1 n2 -- addr count )  >R 64 B/BUF */MOD R> B/SCR * + BLOCK + 64 ;
: .LINE  ( line scr )  (LINE) -TRAILING TYPE ;

*: ."
  34 STATE @ IF COMPILE (.") WORD HERE C@ 1+ ALLOT
  ELSE WORD HERE COUNT TYPE ENDIF ;


: MESSAGE  ( n )
  WARNING @ IF -DUP IF 4 OFFSET @ B/SCR / - .LINE SPACE ENDIF
  ELSE ." MSG # " . ENDIF ;


: (NUMBER)  ( d1 addr1 -- d2 addr2 )
  BEGIN 1+ DUP >R C@ BASE @ DIGIT WHILE
  SWAP BASE @ U* DROP ROT BASE @ U* D+ DPL @ 1+ IF DPL 1+! ENDIF
  R> REPEAT R> ;

;; k8: check this!
: NUMBER  ( addr -- d )
  0 0 ROT DUP 1+ C@ 45 = DUP >R + -1
  BEGIN DPL ! (NUMBER) DUP C@ BL - WHILE DUP C@ 46 - 0 ?ERROR 0 REPEAT
  DROP R> IF DMINUS ENDIF ;

: WORD  ( ch )
  BLK @ IF BLK @ BLOCK ELSE TIB @ ENDIF
  IN @ + SWAP ENCLOSE
  HERE 34 BLANKS
  IN +!
  OVER - >R R HERE C! +
  HERE 1+ R> CMOVE ;

\ : -FIND  ( -- pfa len-byte TRUE // FALSE )
\   BL WORD HERE CONTEXT @ @ (FIND) DUP 0=
\   IF DROP HERE LATEST (FIND) ENDIF ;

;; k8: do not search the same vocab twice
: -FIND  ( -- pfa len-byte TRUE // FALSE )
  BL WORD
  HERE CONTEXT @ @ (FIND) IF TRUE EXIT ENDIF
  CONTEXT @ @ LATEST - 0IF FALSE EXIT ENDIF
  HERE LATEST (FIND) ;

*: '  \ name
  -FIND 0= 0 ?ERROR DROP [COMPILE] LITERAL ;


\ FIXME: this is broken -- it should fix voclink!
: FORGET  \ name
  CURRENT @ CONTEXT @ - 24 ?ERROR
  [COMPILE] ' DUP FENCE @ U< 21 ?ERROR
  DUP NFA (DP) !
  LFA @ CURRENT @ ! ;


: SPACES  ( n )  0 MAX -DUP IF 0 DO SPACE LOOP ENDIF ;
;; new k8 version
\ : SPACES  ( n )  DUP 0<= IF DROP ELSE 1- SPACE RECURSE-TAIL ENDIF ;

: HOLD  ( ch )  -1 HLD +! HLD @ C! ;
: SIGN  ( n d -- d )  ROT 0< IF 45 HOLD ENDIF ;
: <#  PAD HLD ! ;
: #>  ( d -- addr count )  DROP DROP HLD @ PAD OVER - ;
: #   ( d1 -- d2 )  BASE @ M/MOD ROT 9 OVER < IF 7 + ENDIF 48 + HOLD ;
: #S  ( d1 -- d2 )  BEGIN # ( OVER OVER)2DUP OR 0= UNTIL ;

: D.R ( d n )  >R SWAP OVER DABS <# #S SIGN #> R> OVER - SPACES TYPE ;
: .R  ( n1 n2 )  >R S->D R> D.R ;

: D.  ( d )  0 D.R SPACE ;
: .   ( n )  S->D D. ;
: ?   ( addr )  @ . ;
: U.  ( u )  0 D. ;

: U.R ( u n )  >R 0 R> D.R ;


\ <zx-hidden>
\ : ID-TYPE  ( addr count )
\   -DUP IF OVER + SWAP DO I C@ 127 AND EMIT LOOP ELSE DROP ENDIF ;
: ID-TYPE  ( addr count )
  BEGIN DUP WHILE SWAP C@++ 127 AND
    DUP BL < IF DROP [CHAR] ? ENDIF
    EMIT SWAP 1-
  REPEAT 2DROP ;
\ <zx-normal>

;; k8: this was a copying one. rewritten.
: ID.  ( nfa )  IDCOUNT ID-TYPE ;

<zx-hidden>
: WRAP-.ID  ( nfa )
  OUT C@ IF
    DUP C@ 31 AND OUT C@ + $FFE0 AND IF CR OUT OFF ELSE SPACE ENDIF
  ENDIF ID. ;
<zx-normal>

: VLIST
  ENDCR OUT OFF CONTEXT @ @
  BEGIN DUP WRAP-.ID PFA LFA @ DUP 0= ?TERMINAL OR UNTIL DROP ;

<zx-done>
