;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; interpreter and other interactive things
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

;; k8: check this!
: EXPECT  ( addr count )
  OVER + OVER DO  ( addr )
    KEY
    DUP 12 = IF DROP
      DUP I = IF ERR-BEEP 0 -1 ELSE 8 -2 ENDIF
      R> + >R
    ELSE
      DUP 13 = IF LEAVE DROP BL 0
      ELSE DUP BL < IF DROP [CHAR] ? ENDIF DUP ENDIF
      I C! I 1+ OFF
    ENDIF
    DUP IF EMIT ELSE DROP ENDIF
  LOOP DROP ;

: QUERY  TIB @ 80 EXPECT IN OFF ;

\ : ?STACK  SP@ (SP0) @ SWAP U< 1 ?ERROR SP@ HERE 128 + U< 7 ?ERROR ;
: ?STACK  DEPTH 0< 1 ?ERROR SP@ HERE 128 + U< 7 ?ERROR ;

;; special word, name is zero byte.
;; k8: check this!
*: [X00]
  BLK @ IF
    BLK 1+! IN OFF
    BLK @ B/SCR ( 1 -)1- AND 0=
    IF ?EXEC R> DROP ENDIF
  ELSE R> DROP ENDIF ;

;; k8: check this!
: INTERPRET
  BEGIN
    -FIND IF
      STATE @ < IF CFA CALL, ELSE CFA EXECUTE ENDIF
    ELSE
      HERE NUMBER
      DPL @ 1+ IF DLITERAL ELSE DROP LITERAL ENDIF
    ENDIF
  ?STACK AGAIN ;


<zx-hidden>
: (.BASE)
  BASE @ DUP 10 -
  IF [CHAR] [ EMIT DUP DECIMAL 0 .R BASE ! [CHAR] ] EMIT ELSE DROP ENDIF ;

: (.DEPTH)
  DEPTH DUP IF BASE @ DECIMAL [CHAR] ( EMIT SWAP 0 .R BASE ! [CHAR] ) EMIT ELSE DROP ENDIF ;

: (.INV)  ( flag )  20 EMIT 0<> EMIT ;
<zx-normal>

: QUIT
  BLK OFF [COMPILE] [
  BEGIN
    RP!
    ENDCR
    \ 1 (.INV)
    STATE @ IF [CHAR] > EMIT ENDIF [CHAR] > EMIT
    \ 0 (.INV)
    QUERY ENDCR INTERPRET
    STATE @ 0IF ?SPACE 1 (.INV) (.BASE) (.DEPTH) ." ok\x14\x00" ENDIF
  AGAIN ;


: ABORT
  SP! DECIMAL
  CLS
  ." 48K SPECTRUM fig-FORTH 1.1B\n\x7F Abersoft:1983\nmodified by Ketmar Dark, 2024\n\n"
  [COMPILE] FORTH DEFINITIONS QUIT ;

;; basic error will be re-routed here
: (BERR)
  SP! RP! DECIMAL
  ." \x02BERR #" 23610 C@ C>S 1+ . CR
  [COMPILE] FORTH DEFINITIONS QUIT ;

: WARM
  EMPTY-BUFFERS ABORT ;

: COLD
  EMPTY-BUFFERS
  FIRST USE !
  FIRST PREV !
  DR0
  ;; init uservars
  [ zx-cold-uvars^ ] {#,} [ zx-uvars-origin^ ] {#,} [ zx-cold-#uvars ] {#,} CMOVE
  ;; reset DP, FENCE, VOCLINK
  [ zx-dpx-init^ ] {#,} @ (DP) !
  [ zx-dpx-init^ 2 + ] {#,} @ FENCE !
  [ zx-dpx-init^ 4 + ] {#,} @ VOC-LINK !
  ;; reset FORTH-LATEST
  [ zx-dpx-init^ 6 + ] {#,} @ [ zx-forth-vocid ] {#,} !
  ABORT ;


<zx-hidden>
code: (WARM)
zx-word-warm:
  ld    bc, # zx-['pfa] WARM

zx-do-warm:
  ;; setup error handler
  ld    hl, # zx-word-par-baserr
  ld    zx-baserr-stack (), hl
  ld    hl, # zx-baserr-stack
  ld    23613 (), hl

  ld    ix, # zx-uservars-origin
  \ ld    sp, # $CB40 \ FIXME
  \ ld    hl, # $CBE0 \ FIXME
  ld    sp, () zx-cold-uvars
  ld    hl, () zx-cold-uvars 2+
  ld    zx-rp (), hl
  ei
  push  bc
  ld    a, # 2
  call  # $1601
  pop   bc
;code

code: (COLD)
zx-word-cold:
  ld    a, # 8
  ld    $5C6A (), a
  ld    bc, # zx-['pfa] COLD
  jr    # zx-do-warm
;code-no-next

;; return from basic error
code: (BASERR)
zx-word-par-baserr:
  ld    bc, # zx-['pfa] (BERR)
  jr    # zx-do-warm
flush!
  ;; reserve some room for error stack
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw,
zx-baserr-stack:
  0 dw, 0 dw,
;code-no-next

<zx-normal>


<zx-done>
