;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; dictionary and article header creation
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
zx article header:
  db  nlen-flags  ;; name length, flags
    name (with bit 7 set at the last char)
  dw  lfa         ;; link to the previous word
itc:
  dw  cfa
  pfa is here
dtc:
  db  opcode  (JP/CALL) ($C3/$CD)
  dw  dest
  pfa is here

nlen bits:
  bit 7: always set   $80
  bit 6: immediate    $40
  bit 5: smudge       $20

old vocabulary format:
  cfa is "zx-do-does"
  pfa is "zx-does-vocab"
  $A081 -- dummy name
  dw latest-nfa
  dw voc-link

new vocabulary format:
  cfa is "zx-do-vocab"
  dw 0  ; nothing for now
  $A081 -- dummy name
  dw latest-nfa
  dw voc-link

last vocab word is linked to the previous vocab dummy name.
*)

extend-module TCOM

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX utilities

 0 constant zxc-none
-1 constant zxc-code
 1 constant zxc-colon
zxc-none quan zx-compile-mode

$40 constant zx-immediate-mask
$20 constant zx-smudge-mask

$C3 constant zx-jp-opcode
$CD constant zx-call-opcode

OPT-DTC? [IF] 3 [ELSE] 2 [ENDIF] constant zx-#cfa

: zx-nfa>lfa  ( zx-nfa -- zx-lfa )  dup zx-c@ $1F and + 1+ ;
: zx-lfa>nfa  ( zx-lfa -- zx-nfa )  1- << 1- dup zx-c@ $80 < ?^|| else| >> ;

: zx-lfa>cfa  ( zx-lfa -- zx-cfa )  2+ ;
: zx-cfa>lfa  ( zx-lfa -- zx-cfa )  2- ;

: zx-cfa>pfa  ( zx-cfa -- zx-pfa )  zx-#cfa + ;
: zx-pfa>cfa  ( zx-pfa -- zx-cfa )  zx-#cfa - ;

: zx-lfa>pfa  ( zx-lfa -- zx-cfa )  zx-lfa>cfa zx-cfa>pfa ;
: zx-cfa>nfa  ( zx-cfa -- zx-cfa )  zx-cfa>lfa zx-lfa>nfa ;

(*
: zx-pfa>nfa  ( zx-pfa -- zx-nfa TRUE // FALSE )
  5 - dup zx-c@ $80 < ?exit< drop false >?
  << 1- dup zx-c@ $80 < ?^|| else| >> true ;

: zx-.id  ( zx-nfa )
   dup zx-c@ $1F and swap << 1+ dup zx-c@ $7F and emit 1 under- over ?^|| else| 2drop >> ;
*)


: zx-context@  ( -- zx-addr )  zx-context^ zx-w@ ;
: zx-current@  ( -- zx-addr )  zx-current^ zx-w@ ;

: zx-context!  ( value )  zx-context^ zx-w! ;
: zx-current!  ( value )  zx-current^ zx-w! ;

;; set in "zx-mk-header"
0 quan zx-latest-cfa  ;; always valid
0 quan zx-latest-nfa  ;; can be 0 if the current word has no header

: zx-latest-pfa  ( -- zx-pfa )  zx-latest-cfa zx-cfa>pfa ;

|: ?zx-last-nfa  ( -- zx-nfa )
  zx-latest-nfa dup not?error" last defined ZX word is hidden" ;

: zx-immediate  ?zx-last-nfa dup zx-c@ zx-immediate-mask xor swap zx-c! ;
: zx-smudge     zx-latest-nfa dup ?< dup zx-c@ zx-smudge-mask xor swap zx-c! || drop >? ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; compile call to another word, and CFA field

OPT-DTC? [IF]
: zx-resolve-cfa  ( zx-cfa -- zx-cfa )
  dup not?exit  ;; forward
  << dup zx-c@ zx-jp-opcode = ?^| 1+ zx-w@ |? else| >> ;
[ELSE]
: zx-resolve-cfa  ( zx-cfa -- zx-cfa )  ;
[ENDIF]

: zx-call,  ( zx-cfa )  zx-resolve-cfa zx-w, ;

;; compile CFA field.
;; special value: -1 means "code word CFA"
: zx-cfa,  ( zx-addr )
  [ OPT-DTC? ] [IF]
    dup 1+ ?< zx-call-opcode zx-c, zx-w,
    || drop zx-jp-opcode zx-c, zx-here 2+ zx-w, >?
  [ELSE]
    dup 1+ not?< drop zx-here 2+ >? zx-w,
  [ENDIF] ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; host shadow dictionaries

(*
shadow word format:
  dd zx-cfa   ;; -1 means "forward reference"
  dd fix-chain-head  ;; for forward references; 0 for defined words
*)

0
new-field ss>zxcfa
new-field ss>chain
constant #ss

;; fixup chain item
0
new-field fix>next    -- 0: no more
new-field fix>zxaddr  -- zx address
new-field fix>fline   -- source line
new-field fix>fname   -- pointer to file name nfa
constant #fixup

;; file names (for forwards)
module FILES
<case-sensitive>
end-module (private)

module FORTH-SHADOWS
end-module

module EDITOR-SHADOWS
end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; forward references

0 quan last-used-file-nfa  (private)
0 quan last-used-file-uid0  (private)
0 quan last-used-file-uid1  (private)

|: fname-last?  ( uid0 uid1 -- fname-nfa TRUE // FALSE )
  last-used-file-uid1 = swap last-used-file-uid0 = and not?exit&leave
  last-used-file-nfa true ;

|: fname-ptr  ( -- fname-nfa )
  inc-uidx  2dup or not?error" not in include!"
  fname-last? ?exit
  inc-fname
  push-cur voc-cur: files system:mk-create-4
  inc-uidx swap
  dup last-used-file-uid0:! ,
  dup last-used-file-uid1:! ,
  system:latest-nfa pop-cur
  dup last-used-file-nfa:! ;


|: mk-fixup  ( zx-addr chain-head^ )
  fname-ptr >r
  align-here-4 here >r
  ( next) dup @ , r> swap !
  ( zx-addr) lo-word ,
  ( file line) inc-line# ,
  ( file name) r> , ;

|: dump-chain-item  ( chain^ )
  ." CITEM=$" dup .hex8
  ."  NEXT=$" dup fix>next:@ .hex8
  ."  ZXADDR=$" dup fix>zxaddr:@ .hex4
  ."  FLINE=" dup fix>fline:@ base @ decimal swap 0.r base !
  ."  FNAME=" dup fix>fname:@ debug:.id
  drop ;

|: fix-chain  ( zx-addr chain^ )
  dup not?< 2drop exit >?
\ endcr ." FIX-CHAIN: zx-addr=$" over .hex8 ."  chead=$" dup .hex8 cr
  swap >r <<  ( chain | zx-addr )
    dup ?^| \ endcr ."   ITEM:  " dup dump-chain-item cr
            dup fix>zxaddr:@ dup zx-w@ r@ + swap zx-w!
            fix>next:@ |?
  else| drop rdrop >> ;


false quan undefined-forwards? (private)

|: dump-chain-files  ( chain^ )
  << dup ?^|
    endcr ."   used around "
    dup fix>fname:@ debug:.id
    dup fix>fline:@ [char] : emit 0.r cr
  fix>next:@ |?
  else| drop >> ;

|: check-forward  ( shadow-cfa -- res )
\ endcr ." CHECKING: \'" dup dart:cfa>nfa debug:.id
\ ." \': zxcfa=$" dup dart:cfa>pfa ss>zxcfa:@ .hex8
\ ."  chain=$" dup dart:cfa>pfa ss>chain:@ .hex8 cr
  dup dart:cfa>pfa ss>zxcfa:@ 1+ ?exit< drop 0 >?
  dup dart:cfa>pfa ss>chain:@ not?exit< drop 0 >?
  endcr ." TCOM: unresolved forward \'" dup dart:cfa>nfa debug:.id ." \'\n"
  dart:cfa>pfa ss>chain:@ dump-chain-files
  undefined-forwards?:!t 0 ;

@: check-forwards
  undefined-forwards?:!f
  vocid: forth-shadows ['] check-forward vocid-foreach drop
  vocid: editor-shadows ['] check-forward vocid-foreach drop
  undefined-forwards? ?error" TCOM: found some unresolved forwards" ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; shadow words (creation, doers, etc.)

false quan zx-in-editor?

;; zx-vocabulary sets this
0 quan zx-latest-vocid

0 quan zx-forth-vocid
0 quan zx-editor-vocid


: <zx-editor>  zx-in-editor?:!t zx-editor-vocid dup zx-current! zx-context! ;
: <zx-forth>   zx-in-editor?:!f zx-forth-vocid dup zx-current! zx-context! ;


: curr-shadow-vocid  ( -- vocid )
  zx-in-editor? ?< vocid: editor-shadows || vocid: forth-shadows >? ;


: zx-comp!  $C0 zx-state^ zx-w! ;
: zx-exec!  0 zx-state^ zx-w! ;

: zx-comp?  ( -- flag )  zx-state^ zx-w@ 0<> ;
: zx-exec?  ( -- flag )  zx-state^ zx-w@ 0= ;

: zx-?comp  zx-comp? not?error" ZX compilation mode expected" ;
: zx-?exec  zx-exec? not?error" ZX execution mode expected" ;


|: ss-call,  ( pfa )
  dup ss>zxcfa:@ -?< zx-here swap ss>chain:^ mk-fixup 0 || ss>zxcfa:@ >?
  zx-call, ;

;; doers are used to check word kind
: ss-forth-doer     ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-code-doer      ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-variable-doer  ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-2variable-doer ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-uservar-doer   ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-does-doer      ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-vocab-doer     ( pfa )  system:?exec zx-?comp ss-call, ;

: ss-constant-doer ( pfa )
  zx-comp? ?exit< system:?exec ss-call, >?
  ss>zxcfa:@ zx-cfa>pfa zx-w@ [\\] {#,} ;

: zx-checker:  ( shadow-cfa )  \ name
  <builds ,
 does> ( cfa pfa )
  over system:does? not?exit< 2drop false >?
  @ swap system:doer@ = ;

['] ss-forth-doer zx-checker: zx-forth?
['] ss-code-doer zx-checker: zx-code?
['] ss-constant-doer zx-checker: zx-constant?
['] ss-variable-doer zx-checker: zx-variable?
['] ss-2variable-doer zx-checker: zx-2variable?
\ ['] ss-quan-mth::zx-quan-doer zx-checker: zx-quan?
['] ss-uservar-doer zx-checker: zx-uservar?
['] ss-vocab-doer zx-checker: zx-vocab?
['] ss-does-doer zx-checker: zx-does?


;; define undefined forward value
|: ss-define-forward  ( zx-addr shadow-cfa )
  over hi-word ?error" wut?!"  -- just in case
  dup dart:cfa>pfa ss>zxcfa:@ +0?<
    " duplicate definition of forward \'" string:>pad
    dart:cfa>nfa idcount string:pad+cc
    " \'!" string:pad+cc
    string:pad-cc@ error >?
\ endcr ." *** FIX FORWARD \'" dup dart:cfa>nfa debug:.id ." \'\n"
  dart:cfa>pfa
  2dup ss>zxcfa:!
  ss>chain:@ fix-chain ;

|: check-existing-shadow  ( zx-cfa addr count -- zx-cfa addr count FALSE // TRUE )
  2dup curr-shadow-vocid vocid-find not?exit&leave
  dup ss>zxcfa:@ +0?< drop
    " duplicate definition of forward \'" string:>pad
    string:pad+cc " \'!" string:pad+cc
    string:pad-cc@ error >?
  nrot 2drop  ;; drop addr and count
  over ?< ss-define-forward || 2drop >?
  true ;

: (mk-shadow-header)  ( zx-cfa addr count doer )
  >r check-existing-shadow ?exit< rdrop >?
  push-cur curr-shadow-vocid current! system:mk-builds
  ( zx-cfa) ,
  ( chain) 0 ,
  system:latest-cfa pop-cur r> system:!doer ;

: mk-shadow-header  ( addr count doer )
  >r zx-here nrot r> (mk-shadow-header) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create ZX word header

true quan zx-headers-allowed?

: <zx-hidden>  zx-headers-allowed?:!f ;
: <zx-normal>  zx-headers-allowed?:!t ;

|: error-bad-name  ( addr count )
  0 max 1024 min
  " invalid ZX word name: \'" string:>pad string:pad+cc
  " \'!" string:pad+cc
  string:pad-cc@ error ;

|: ?good-zx-name  ( addr count )
  dup 0 32 within not?< error-bad-name >?
  2dup swap <<
    over ?^| c@++ 32 128 within not?< 2drop error-bad-name >?
            1 under- |?
  else| 4drop >> ;

;; up to CFA.
;; "[X00]" is the special name -- word with zero name byte. yes, it is used.
: zx-mk-header  ( addr count )
  2dup " [X00]" string:=ci ?< 2drop " \x00" || 2dup ?good-zx-name >?
  zx-headers-allowed? not?exit< 2drop
    zx-here zx-latest-cfa:!
    zx-latest-nfa:!0 >?
  zx-here zx-latest-nfa:!
\ endcr ." ZX: new header \'" 2dup type ." \' at $" zx-here .hex4 cr
  ;; name
  dup $80 or zx-c,
  swap << over ?^| c@++ string:upchar zx-c, 1 under- |? else| 2drop >>
  zx-here 1- dup zx-c@ $80 or swap zx-c!
\ endcr ." 102: last name byte at $" zx-here 1- .hex4 cr
  ;; lfa
  zx-current@ zx-w@ zx-w,
  zx-latest-nfa zx-current@ zx-w!
  zx-here zx-latest-cfa:! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; define some forwards

-1 " LIT" ['] ss-code-doer (mk-shadow-header)
OPT-CLIT? [IF]
-1 " CLIT" ['] ss-code-doer (mk-shadow-header)
-1 " BLIT" ['] ss-code-doer (mk-shadow-header)
[ENDIF]
-1 " BRANCH" ['] ss-code-doer (mk-shadow-header)
-1 " 0BRANCH" ['] ss-code-doer (mk-shadow-header)
-1 " TBRANCH" ['] ss-code-doer (mk-shadow-header)
|: (name-x0)  " (\')" ;  |: (name-x1)  " (.\')" ;
-1 (name-x0) ['] ss-code-doer (mk-shadow-header)
-1 (name-x1) ['] ss-code-doer (mk-shadow-header)
-1 " ;" ['] ss-code-doer (mk-shadow-header)
-1 " ;S" ['] ss-code-doer (mk-shadow-header)
-1 " COMPILE" ['] ss-code-doer (mk-shadow-header)

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; compile zx literals

: zx-#,  ( n )
  dup -32768 65536 within not?error" ZX literal out of bounds"
  [ OPT-CLIT? ] [IF]
    dup hi-byte <<
        0 of?v| forth-shadows:CLIT zx-c, |?
      255 of?v| forth-shadows:BLIT zx-c, |?
    else| drop forth-shadows:LIT zx-w, >>
  [ELSE]
    forth-shadows:LIT zx-w,
  [ENDIF] ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; branch chain compilers/resolvers

OPT-REL-BRANCH? [IF]
: zx-brdest@    ( zx-addr )  dup zx-w@ + lo-word ;
: zx-brdest!    ( zx-dest zx-addr )  tuck - swap zx-w! ;
[ELSE]
: zx-brdest@    ( zx-addr )  zx-w@ ;
: zx-brdest!    ( zx-dest zx-addr )  zx-w! ;
[ENDIF]

: zx-brdest,  ( zx-dest )  zx-here 0 zx-w, zx-brdest! ;

: zx-<j-mark    ( -- zx-dest )  zx-here ;
: zx-chain-j>   ( chain -- chain-new )  zx-here swap zx-w, ;
: zx-mark-j>    ( -- chain-new )  0 zx-chain-j> ;
: zx-<j-resolve ( zx-dest )  zx-brdest, ;

: zx-resolve-j> ( chain )
  dup not?exit< drop >? zx-here >r
  << dup zx-w@ r@ rot zx-brdest! dup ?^|| else| drop rdrop >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helper vocab with tcom words
;; this also compiles literals

: zx-?in-colon
  system:?exec zx-?comp
  zx-compile-mode zxc-colon = not?error" ZX colon word was not started yet" ;


module SHADOW-HELPERS
*: (  [\\] forth:( ;
*: \  [\\] forth:\ ;
*: -- [\\] forth:-- ;
*: // [\\] forth:// ;
*: (* [\\] forth:(* ;
*: (+ [\\] forth:(+ ;

*: [IF]     [\\] forth:[IF] ;
*: [IFNOT]  [\\] forth:[IFNOT] ;
*: [ELSE]   [\\] forth:[ELSE] ;
*: [ENDIF]  [\\] forth:[ENDIF] ;

*: RECURSE-TAIL
  zx-?in-colon
  forth-shadows:BRANCH zx-latest-pfa zx-brdest, ;

*: RECURSE
  zx-?in-colon
  zx-latest-cfa zx-call, ;

*: FALSE  zx-?in-colon  0 zx-#, ;
*: TRUE   zx-?in-colon  1 zx-#, ;

*: [CHAR]
  zx-?in-colon
  parse-name 1 <> ?error" [CHAR] expects a char"
  c@ zx-#, ;

*: [COMPILE]
  zx-?in-colon
  parse-name
  2dup curr-shadow-vocid vocid-find not?<
    " ZX word \'" string:>pad string:pad+cc
    " \' not found!" string:pad+cc
    string:pad-cc@ error >?
  nrot 2drop execute ;

;; sadly, i have to made "COMPILE" like this.
;; the original wasn't even an immediate word.
*: COMPILE
  zx-?in-colon
  parse-name
  2dup curr-shadow-vocid vocid-find not?<
    " ZX word \'" string:>pad string:pad+cc
    " \' not found!" string:pad+cc
    string:pad-cc@ error >?
  forth-shadows:compile execute 2drop ;

*: ."
  zx-?in-colon
  34 parse-qstr
  forth-shadows:(.") zx-bstr, ;

|: (mk#)  ( value )
  zx-?in-colon zx-#, ;

*: [ZX-DO-COLON]     zx-do-colon^ (mk#) ;
*: [ZX-DO-CONSTANT]  zx-do-constant^ (mk#) ;
*: [ZX-DO-2CONSTANT] zx-do-2constant^ (mk#) ;
*: [ZX-DO-VARIABLE]  zx-do-variable^ (mk#) ;
*: [ZX-DO-2VARIABLE] zx-do-2variable^ (mk#) ;
*: [ZX-DO-USERVAR]   zx-do-uservar^ (mk#) ;
*: [ZX-DO-VOCAB]     zx-do-vocab^ (mk#) ;
*: [ZX-DO-DOES]      zx-do-does^ (mk#) ;

end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX code words support vocab

module ZXF-CODE
<disable-hash>

extend-module z80asm:instr
;; macro
: next
  z80asm:flush
  z80asm:postfix? >r
  true z80asm:postfix!
  " i-next" z80-labman:use-label # jp
  r> z80asm:postfix! ;
: next-push-hl
  z80asm:flush
  z80asm:postfix? >r
  true z80asm:postfix!
  " i-next-push-hl" z80-labman:use-label # jp
  r> z80asm:postfix! ;
: next-push-de-hl
  z80asm:flush
  z80asm:postfix? >r
  true z80asm:postfix!
  " i-next-push-de-hl" z80-labman:use-label # jp
  r> z80asm:postfix! ;
end-module

: ;code-no-next
  zx-compile-mode zxc-code = not?error" ZX code word was not started yet"
  z80asm:flush
  zx-fix-dp
  pop-ctx [\\] zx-code-def:<end-asm>
  zxc-none zx-compile-mode:! ;

: ;code
  z80asm:instr:next
  ;code-no-next ;

: ;code-push-hl
  z80asm:instr:next-push-hl
  ;code-no-next ;

: ;code-push-de-hl
  z80asm:instr:next-push-de-hl
  ;code-no-next ;

end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create ZX code word

|: define-zx-code-word-label  ( addr count )
  " zx-word-" string:>pad string:pad+cc
  string:pad-cc@ zx-here z80-labman:lbl-new
  z80-labman:reset-locals ;

: zx-code:  \ name
  zx-compile-mode ?error" still compiling something"
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r@ ['] ss-code-doer mk-shadow-header
  ( cfa) -1 zx-cfa,
  zx-fix-org
  2r> define-zx-code-word-label
  [\\] <asm>
  push-ctx voc-ctx: zxf-code
  zxc-code zx-compile-mode:! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; separate wordlist for ";", to make helper definitions easier
module SHADOW-SEMI
<disable-hash>

*: [
  system:?exec zx-?comp
  zx-compile-mode zxc-colon = not?error" invalid \'[\' usage in ZX mode"
  push-ctx
  \ -666 vsp-push
  \ voc-ctx: forth
  0 vsp-push
  vocid: forth vsp-push
  vocid: shadow-semi vsp-push
  voc-ctx: tcom
  zx-exec! ;

*: {#,}
  system:?exec zx-?comp
  zx-compile-mode zxc-colon = not?error" invalid \'{#,}\' usage in ZX mode"
  zx-#, ;

*: ]
  system:?exec zx-?exec
  zx-compile-mode zxc-colon = not?error" invalid \']\' usage in ZX mode"
  \ << vsp-pop -666 <> ?^|| else| >> pop-ctx
  << vsp-pop ?^|| else| >> pop-ctx
  zx-comp! ;

*: ;
  system:?exec zx-?comp
  $c0de_0800 system:?pairs
  forth-shadows:;s
  zx-smudge
  zx-exec!
  zx-compile-mode zxc-colon = not?error" ZX colon word was not started yet"
  zxc-none zx-compile-mode:!
  << vsp-pop ?^|| else| >> pop-ctx ;
end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create colon words

: zx:  \ name
  zx-compile-mode ?error" still compiling something"
  system:?exec zx-?exec
  parse-name
\ endcr ." 000: DP=" zx-here .hex4
\ ."  LATEST-LFA=$" zx-latest-lfa .hex4
\ ."  CURRENT=$" zx-current^ .hex4
\ ."  [CURRENT]=$" zx-current^ zx-w@ .hex4
\ ."  [CURRENT]=$" zx-current@ .hex4
\ cr
  2dup 2>r zx-mk-header
\ endcr ." 010\n"
\ endcr ."   LATEST-LFA=$" zx-latest-lfa .hex4 ."  LATEST-NFA=$" zx-latest-lfa zx-lfa>nfa .hex4 cr
\ endcr zx-latest-lfa 1- zx-c@ ."  [-1]=$" .hex2 cr
  zx-smudge
  2r> ['] ss-forth-doer mk-shadow-header
  ( cfa) zx-do-colon^ zx-cfa,
  zx-fix-org
  push-ctx
  0 vsp-push
  vocid: forth-shadows vsp-push
  zx-in-editor? ?< vocid: editor-shadows vsp-push >?
  vocid: shadow-helpers vsp-push
  voc-ctx: shadow-semi
  zx-comp!
  $c0de_0800
  zxc-colon zx-compile-mode:! ;

: *zx:  \ name
  [\\] zx: zx-immediate ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create constants, vars, etc.

: zx-constant  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-constant-doer mk-shadow-header
  ( cfa) zx-do-constant^ zx-cfa,
  ( value) zx-w,
  zx-fix-org ;

: zx-variable  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-variable-doer mk-shadow-header
  ( cfa) zx-do-variable^ zx-cfa,
  ( value) zx-w,
  zx-fix-org ;

: zx-ud-buffer  ( size )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-variable-doer mk-shadow-header
  ( cfa) zx-do-variable^ zx-cfa,
  zx-allot
  zx-fix-org ;

: zx-2variable  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-2variable-doer mk-shadow-header
  ( cfa) zx-do-2variable^ zx-cfa,
  ( value) dup zx-cfa, hi-word zx-cfa,
  zx-fix-org ;

: zx-uservar  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-uservar-doer mk-shadow-header
  ( cfa) zx-do-uservar^ zx-cfa,
  2 zx-uservar-nallot
  ( offset) dup zx-uservar-start^ - zx-c,
  zx-w! ;; set value
  zx-fix-org ;

: zx-uvar-ofs  ( -- value ) \ name
  parse-name vocid: forth-shadows vocid-find not?error" unknown uservar"
  dup zx-uservar? not?error" not a uservar"
  dart:cfa>pfa ss>zxcfa:@ dup not?error" undefined uservar"
  zx-cfa>pfa zx-c@ ;

: zx-vocabulary  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-vocab-doer mk-shadow-header
  ( cfa) zx-do-vocab^ zx-cfa,
  ( reserved) 0 zx-w,
  ( dummy name) $A081 zx-w,
  zx-here zx-latest-vocid:!
  ( latest-nfa) zx-current@ 2- zx-w,
  ( voc-link) zx-here zx-voclink^ zx-w@ zx-w, zx-voclink^ zx-w!
  zx-fix-org ;


|: (-zx-find)  ( -- zx-cfa )
  parse-name
  2dup curr-shadow-vocid vocid-find not?<
    vocid: forth-shadows vocid-find not?error" unknown ZX word"
  || nrot 2drop >?
  dart:cfa>pfa ss>zxcfa:@ dup not?error" undefined word" ;

*: zx-[']  ( -- zx-cfa )  \ name
  (-zx-find) [\\] {#,} ;

*: zx-['pfa]  ( -- zx-cfa )  \ name
  (-zx-find) zx-cfa>pfa [\\] {#,} ;


module ZX-DEFS
<disable-hash>
*: constant  ( value )  \ name
  [\\] zx-constant ;

*: variable  ( value )  \ name
  [\\] zx-variable ;

*: 2variable  ( value )  \ name
  [\\] zx-2variable ;

*: uservar  ( value )  \ name
  [\\] zx-uservar ;

*: <zx-done>
  system:?exec zx-?exec
  context@ vocid: zx-defs = not?error" unfinished contexts"
  pop-ctx pop-ctx ;

*: code:  \ name
  [\\] zx-code: ;

*: :  \ name
  [\\] zx: ;

*: *:  \ name
  [\\] *zx: ;

end-module

: <zx-definitions>
  push-ctx voc-ctx: tcom
  push-ctx voc-ctx: zx-defs
  <zx-normal> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SHADOWS vocab methods (compile literals, create forwards)

(*
: zx-voc-execcomp  ( cfa -- ... TRUE // FALSE )
\ endcr dup zx-cfa>nfa ." ZX: |" zx-.id ." |\n"
  zx-?comp
  dup zx-cfa>nfa zx-c@ zx-immediate-mask and ?<
    endcr ." cannot compile immediate ZX word \'" zx-cfa>nfa zx-.id ." \'!\n"
    abort >?
  zx-w, true ;
['] zx-voc-execcomp vocid: zxw system:vocid-execcomp-cfa!
*)

(*
: zx-voc-finder  ( addr count skip-hidden? vocid -- cfa-xt TRUE // FALSE )
  2drop zx-find ;
['] zx-voc-finder vocid: zxw system:vocid-find-cfa!
*)

|: zx-voc-literal  ( lit -- ... TRUE // lit FALSE )
  zx-comp? not?exit&leave
  system:?exec zx-#, true ;
['] zx-voc-literal vocid: forth-shadows system:vocid-literal-cfa!

|: zx-voc-notfound  ( addr count -- processed-flag )
\ endcr ." ZX-NOTFOUND: |" 2dup type ." |\n"
  zx-comp? not?exit< 2drop false >?
  ;; create forward reference
  -1 nrot ['] ss-forth-doer (mk-shadow-header)
  curr-shadow-vocid system:vocid-latest dart:lfa>pfa ss-call,
  true ;
['] zx-voc-notfound vocid: forth-shadows system:vocid-notfound-cfa!

;; create "FORTH" vocabulary word
: create-first-zx-word
  system:?exec zx-?exec
  zx-current^ not?error" ZX CURRENT is not defined"
  zx-context^ not?error" ZX CONTEXT is not defined"
  zx-voclink^ not?error" ZX VOCLINK is not defined"
  zx-here 64 + dup zx-current! zx-context!  ;; temporary
  0 zx-current@ zx-w!
  zx-here >r  ;; save NFA
  " FORTH" 2dup zx-mk-header
  ['] ss-vocab-doer mk-shadow-header
  \ ( cfa) zx-do-does^ zx-w,
  \ ( doer) zx-fix-org " zx-doer-vocab" z80-labman:ref-label,
  ( cfa) zx-do-vocab^ zx-cfa,
\ endcr ." FORTH VOCAB PFA=$" zx-here .hex4 cr
  ( reserved) 0 zx-w,
  ( dummy name) $A081 zx-w,
  zx-here zx-forth-vocid:!
\ endcr ." FORTH VOCAB VOCID=$" zx-forth-vocid .hex4 cr
  ( latest-nfa) zx-here dup zx-current! zx-context! r> zx-w,
  ( voc-link) zx-here 0 zx-w, zx-voclink^ zx-w!
  ( for some reason FORTH is immediate) zx-immediate ;

end-module


alias-for tcom:<zx-definitions> is <zx-definitions>
