;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; and now for something completely different...
;; UrForth/Beast: Devastator -- Seductive and Deadly Forth System
;; Copyright (C) 2023-2024 Ketmar Dark // Invisible Vector
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; simple CLI arguments parsing
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
$use <getarg>

getarg:arg: --arg1  endcr ." ARG1!\n" ;
getarg:help:  ." some argument" ;

getarg:arg: --do-arg2  ( addr count )  endcr ." ARG2! " type ;
getarg:help:  ." some other argument" ;
getarg:last-need-arg

getarg:last-alias: -a2

;; optional
getarg:usage: ." usage: runme [args] fname\nargs:\n" ;

;; optional
getarg:file: ( addr count )
  endcr ." file: \'" type ." \'\n" ;

getarg:parse-args
*)

module getarg
<disable-hash>
<published-words>

;; current arg#, used in "parse-args".
;; it will be advanced before calling an argument handler.
0 quan cur-arg#

;; used in "parse-args"
false quan seen-ddash?

;; vocab with CLI arg words
0 quan args-vocid

;; called for unknown argument; default is abort
vect file-arg  ( addr count )

;; show usage (with type)
vect-empty usage

;; add arg words there
module args
<disable-hash>
end-module
vocid: args args-vocid:!


;; arg word pfa:
;;   processor-cfa  ( -- )
;;   help-cfa       ( -- ) -- print help text; no last cr!
;;   need-arg?

|: arg-doer  ( pfa )  @execute-tail ;

|: arg-word?  ( cfa -- flag )
  dup system:does? not?exit< drop false >?
  system:doer@ ['] arg-doer = ;

|: help@  ( cfa -- help-cfa TRUE // FALSE )
  dup arg-word? not?exit< drop false >?
  dart:cfa>pfa 4+ @ dup not?exit true ;

|: need-arg@  ( cfa -- need-arg? )
  dup arg-word? not?exit< drop false >?
  dart:cfa>pfa 8 + @ ;

: mk-arg  ( proc-cfa addr count )
  2dup args-vocid find-in-vocid ?< drop
    " duplicate CLI arg: \'" string:>pad string:pad+cc [char] " string:pad+char
    (sp0!) string:pad-cc@ error >?
  push-cur args-vocid current!
  system:mk-builds
  ( proc-cfa) ,
  ( help-cfa) 0 ,
  ( need-arg) 0 ,
  system:latest-cfa ['] arg-doer system:!doer
  pop-cur ;

|: last-arg-pfa  ( -- pfa )  push-cur args-vocid current! system:latest-pfa pop-cur ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; argument definition words

: arg:  \ name
  0 parse-name mk-arg
  last-arg-pfa [\\] :noname >r swap ! r> ;
;; help for the last argument
: help:
  last-arg-pfa 4+ [\\] :noname >r swap ! r> ;

: last-need-arg
  true last-arg-pfa 8 + ! ;

: last-alias:  \ name
  last-arg-pfa dup >r @ parse-name mk-arg
  ( copy needarg flag) r> 8 + @ last-arg-pfa 8 + ! ;

: usage:  [\\] :noname swap usage:! ;
: file:   [\\] :noname swap file-arg:! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; help helpers

0 quan max-alen (private)


|: one-arg-len  ( cfa -- res )
  dup help@ not?exit< drop false >? drop
  dup need-arg@ ?< 6 || 0 >? >r
  dart:cfa>nfa idcount r> + max-alen max max-alen:!
  drop false ;

|: collect-one-help  ( cfa -- cfa res )
  dup help@ not?< drop >? false ;

|: print-one-help  ( cfa -- res )
  dup help@ ?< endcr
    swap dup need-arg@ >r dart:cfa>nfa idcount dup nrot type r>
    ( help-cfa nlen need-arg? )
    ?< ."  <arg>" 6 + >?
    max-alen swap - bl #emit
    ."  -- " execute ( type) endcr
  || drop >?
  false ;

|: calc-max-arg-len  max-alen:!0 args-vocid ['] one-arg-len vocid-foreach drop ;
|: collect-args  ( -- 0 ... )  0 args-vocid ['] collect-one-help vocid-foreach drop ;

: show-args-help
  calc-max-arg-len
  collect-args
  << dup ?^| print-one-help drop |? else| drop >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; default "unknown argument" word

: bad-arg  ( addr count )
  " unknown CLI arg: \'" string:>pad string:pad+cc [char] " string:pad+char
  (sp0!) string:pad-cc@ error ;
['] bad-arg file-arg:!


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; default CLI arg parser

;; no checks!
|: (next-arg)  ( -- addr count )  cur-arg# cur-arg#:1+! argv@ ;

|: get-next-arg  ( aaddr acount -- addr count )
  cur-arg# #arg >= ?<
    " missing argument for \'" string:>pad string:pad+cc [char] " string:pad+char
    string:pad-cc@ error >?
  2drop (next-arg) ;

|: do-arg-one-arg  ( addr count cfa )
  dup need-arg@  ( addr count cfa need-arg? )
  ?< >r get-next-arg r> || nrot 2drop >?
  execute-tail ;

: one-arg  ( addr count )
  2dup args-vocid find-in-vocid not?exit< file-arg >?
  dup arg-word? ?< do-arg-one-arg || nrot 2drop execute-tail >? ;

|: parse-files
  << cur-arg# #arg < ?^| (next-arg) file-arg |? else| >> ;

: parse-args
  cur-arg#:!0 seen-ddash?:!f
  << cur-arg# #arg < ?^| (next-arg) one-arg seen-ddash? ?exit< parse-files >? |? else| >> ;

;; use this word in args which require more args ;-)
;; like "--out <name>"
: next-arg  ( errmsg-addr errmsg-count -- addr count )
  cur-arg# #arg < not?error (next-arg) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some predefined arguments

arg: --
  seen-ddash?:!t ;

arg: --help  usage show-args-help (sp0!) bye ;
help: ." show this help" ;
last-alias: -h

seal-module
end-module
