;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; and now for something completely different...
;; UrForth/Beast: Devastator -- Seductive and Deadly Forth System
;; Copyright (C) 2023-2024 Ketmar Dark // Invisible vect
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


here  \ for size reports

module z80dis
<published-words>

0 quan pc  ;; can be bigger than $FFFF
true quan hex?

vect zx-c@  ( pc -- byte )


<private-words>

0 quan (ixiy)  \ contains 0, [char] x or [char] y
0 quan (disp)
0 quan (dptr)
0 quan (opcode)

create (ixydisp-table)
  $00 c, $00 c, $00 c, $00 c,
  $00 c, $00 c, $70 c, $00 c,
  $40 c, $40 c, $40 c, $40 c,
  $40 c, $40 c, $BF c, $40 c,
  $40 c, $40 c, $40 c, $40 c,
  $40 c, $40 c, $40 c, $40 c,
  $00 c, $08 c, $00 c, $00 c,
  $00 c, $00 c, $00 c, $00 c,
create;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

create disbuf 32 allot create;

: (put-char)  ( ch )  (dptr) c! (dptr):1+! ;
: (put-str)  ( addr len )  swap << over +?^| c@++ (put-char) 1 under- |? else| 2drop >> ;

: (put-space)  bl (put-char) ;
: (put-comma)  [char] , (put-char) bl (put-char) ;
: (put-lpar)   [char] ( (put-char) ;
: (put-rpar)   [char] ) (put-char) ;


: (put-str-2)  ( addr )  2 (put-str) ;
: (put-str-4)  ( addr )  drop + 4 (put-str) ;
: (put-str-4x) ( addr )  dup 3 + c@ bl = ?< 3 || 4 >? (put-str) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
: (put-n8)  ( n )
  lo-byte
  base @ >r
  hex? ?< hex <# # # [char] $ #hold || decimal <# #s >?
  #> (put-str) r> base ! ;

: (put-n16)  ( n )
  lo-word
  base @ >r
  hex? ?< hex <# # # # # [char] $ #hold || decimal <# #s >?
  #> (put-str) r> base ! ;

: (put-disp)
  (disp) not?exit
  (disp) -?< [char] - || [char] + >? (put-char)
  (disp) abs
  base @ >r decimal <# #s #>
  (put-str) r> base ! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; advances PC

: (fetch-byte)  ( -- b )  pc lo-word zx-c@ pc:1+! ;
: (fetch-word)  ( -- b )  (fetch-byte) (fetch-byte) 8 lshift or ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
: (put-ixy-mem)
  " (i" (put-str) (ixiy) (put-char) (put-disp) (put-rpar) ;

|: (put-r8hlm)
  (ixiy) ?< (put-ixy-mem) || " (hl)" (put-str) >? ;

;; undocumented IX/IY 8-bit part access
|: (put-r8ixy)
  (ixiy) ?<
    dup [char] h = over [char] l = or ?< (ixiy) (put-char) >?
  >? (put-char) ;

: (put-r8)  ( r8 )
  7 and  " bcdehl.a" drop + c@
  dup [char] . = ?< drop (put-r8hlm) || (put-r8ixy) >? ;

: (put-r16-hl-ixy)
  (ixiy) ?exit< [char] i (put-char) (ixiy) (put-char) >? " hl" (put-str) ;

: (put-v16)  (fetch-word) (put-n16) ;
: (put-m16)  (put-lpar) (put-v16) (put-rpar) ;

: (put-r16-common)  ( r16 addr count )
  drop swap 3 and 2* +
  dup c@ [char] h = ?exit< drop (put-r16-hl-ixy) >? (put-str-2) ;

: (put-r16-sp)  ( r16 )
  (opcode) 4 rshift " bcdehlsp" (put-r16-common) ;

: (put-r16-af)  ( r16 )
  (opcode) 4 rshift " bcdehlaf" (put-r16-common) ;

: (put-cc)  ( cc )
  7 and 2* " nzz ncc popep m " drop + dup c@ (put-char)
  1+ c@ dup 32 <> ?exit< (put-char) >? drop ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
: (decode-cb-unixy)
  ;; special undocumented thing
  (ixiy) not?exit ;; `bit` doesn't need undoc ixy
  (opcode) dup $80 and swap 7 and 6 <> and
  ?< (put-comma) (put-ixy-mem) >? ;

: (decode-cb-bit)
  (opcode) 4 rshift $0C and 4- " bit res set " (put-str-4)
  (put-space)
  (opcode) 3 rshift 7 and [char] 0 + (put-char)
  (put-comma) ;

: (decode-cb-shift)
  (opcode) 2/ $1C and
  " rlc rrc rl  rr  sla sra sll srl " (put-str-4)
  (put-space) ;

: (decode-cb)
  (opcode) $C0 and ?< (decode-cb-bit) || (decode-cb-shift) >?
  (opcode) (put-r8) (decode-cb-unixy) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
: (decode-ed-xrep)
  (opcode) <<
    ;; two instructions with the wrong mnemonic length ;-)
    $A3 of?v| " outi" (put-str) |?
    $AB of?v| " outd" (put-str) |?
  else| ;; common code
    dup 3 and 2* " ldcpinot" drop + (put-str-2)
    dup $08 and ?< [char] d || [char] i >? (put-char)
    $10 and ?< [char] r (put-char) >?
  >> ;

|: (decode-ed-x04-ret)
  " ret" (put-str)
  (opcode) $08 and ?< [char] i || [char] n >?
  (put-char) ;

|: (decode-ed-x04-im)
  " im   " (put-str)
  (opcode) $47 = ?exit< " 0/1" (put-str) >?
  (opcode) $10 and ?<
    (opcode) $08 and ?< [char] 2 || [char] 1 >?
  || [char] 0 >?
  (put-char) ;

|: (decode-ed-x04-07)
  (opcode) <<
    $47 of?v| " ld   i, a" |?
    $4F of?v| " ld   r, a" |?
    $57 of?v| " ld   a, i" |?
    $5F of?v| " ld   a, r" |?
    $67 of?v| " rrd" |?
    $6F of?v| " rld" |?
  else| drop " nope" >>
  (put-str) ;

|: (decode-ed-x04)
  (opcode) 7 and <<
    $04 of?v| " neg" (put-str) |?
    $05 of?v| (decode-ed-x04-ret) |?
    $06 of?v| (decode-ed-x04-im) |?
    $07 of?v| (decode-ed-x04-07) |?
  else| drop " nope" (put-str) >> ;

|: (decode-ed-x02-torr)  (put-r16-sp) (put-comma) (put-m16) ;
|: (decode-ed-x02-tomem) (put-m16) (put-comma) (put-r16-sp) ;

|: (decode-ed-x02)
  ;; r16
  (opcode) $01 and ?exit<
    " ld   " (put-str)
    (opcode) $08 and ?< (decode-ed-x02-torr) || (decode-ed-x02-tomem) >?
  >?
  (opcode) 2/ 4 and " sbc adc " (put-str-4)
  "  hl, " (put-str)
  (put-r16-sp) ;

|: (decode-ed-x01)
  " out  (c), " (put-str)
  (opcode) 3 rshift
  ;; check for `(hl)`, it is special here
  dup 7 and 6 = ?< drop [char] 0 (put-char) || (put-r8) >? ;

|: (decode-ed-x00)
  " in   " (put-str)
  (opcode) 3 rshift
  ;; check for `(hl)`, it is special here
  dup 7 and 6 <> ?< (put-r8) (put-comma) || drop >?
  " (c)" (put-str) ;

: (decode-ed)
  (opcode) $A4 and $A0 = ?exit< (decode-ed-xrep) >?
  (opcode) $C0 and $40 <> ?exit< " nope" (put-str) >?
  (opcode) $04 and ?exit< (decode-ed-x04) >?
  (opcode) $02 and ?exit< (decode-ed-x02) >?
  (opcode) $01 and ?exit< (decode-ed-x01) >?
  (decode-ed-x00) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ld r8,r8 (and halt)
|: (decode-norm-grp1-lixy)
  ;; this is for "ld l,(ix)" and such
  (ixiy) >r
  (opcode) 3 rshift 7 and 6 <> ?< (ixiy):!0 >?
  (opcode) 3 rshift (put-r8)
  (put-comma)
  (opcode) 7 and 6 <> ?< rdrop 0 || r> >? (ixiy):!
  (opcode) (put-r8) ;

: (decode-norm-grp1)
  (opcode) $76 = ?exit< " halt" (put-str) >?
  " ld   " (put-str)
  (ixiy) 0<>  (opcode) 3 rshift 7 and 6 =  (opcode) 7 and 6 = or and ?exit< (decode-norm-grp1-lixy) >?
  (opcode) 3 rshift (put-r8)
  (put-comma)
  (opcode) (put-r8) ;

: (put-alu-str)
  (opcode) 2/ $1C and " add adc sub sbc and xor or  cp  " (put-str-4)
  (put-space)
  ;; two special opcodes
  (opcode) $38 and dup $08 = over $18 = or swap $00 = or
  ?exit< " a, " (put-str) >? ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
\ call,ret,push,pop,etc.
|: (decode-norm-grp3-01)
  (opcode) $08 and ?<
    (opcode) $30 and <<
      $00 of?v| " ret" (put-str) |?
      $10 of?v| " exx" (put-str) |?
      $20 of?v| " jp   (" (put-str) (put-r16-hl-ixy) (put-rpar) |?
      $30 of?v| " ld   sp, " (put-str) (put-r16-hl-ixy) |?
    else| drop >>
  || " pop  " (put-str) (put-r16-af) >? ;

|: (decode-norm-grp3-03)
  (opcode) $38 and <<
    $00 of?v| " jp   " (put-str) (put-v16) |?
    ;; CB:$08 of?v||
    $10 of?v| " out  (" (put-str) (fetch-byte) (put-n8) " ), a" (put-str) |?
    $18 of?v| " in   a, (" (put-str) (fetch-byte) (put-n8) (put-rpar) |?
    $20 of?v| " ex   (sp), " (put-str) (put-r16-hl-ixy) |?
    $28 of?v| " ex   de, hl" (put-str) |?
    $30 of?v| " di" (put-str) |?
    $38 of?v| " ei" (put-str) |?
  else| drop >> ;

|: (decode-norm-grp3-05)
  (opcode) $08 and ?<  ;; prefixes already done, so only CALL is left
    " call " (put-str) (put-v16)
  || " push " (put-str) (put-r16-af) >? ;

: (decode-norm-grp3)
  (opcode) 7 and <<
    $00 of?v| " ret  " (put-str)  (opcode) 3 rshift (put-cc) |?
    $01 of?v| (decode-norm-grp3-01) |?
    $02 of?v| " jp   " (put-str)  (opcode) 3 rshift (put-cc) (put-comma) (put-v16) |?
    $03 of?v| (decode-norm-grp3-03) |?
    $04 of?v| " call " (put-str)  (opcode) 3 rshift (put-cc) (put-comma) (put-v16) |?
    $05 of?v| (decode-norm-grp3-05) |?
    $06 of?v| (put-alu-str) (fetch-byte) (put-n8) |?
    $07 of?v| " rst  " (put-str) (opcode) $38 and (put-n8) |?
  else| drop >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
|: (decode-norm-grp0-00-01)
  (opcode) $08 and ?<
    " add  " (put-str) (put-r16-hl-ixy)
    (put-comma) (put-r16-sp)
  ||
    " ld   " (put-str) (put-r16-sp)
    (put-comma) (put-v16) >? ;

|: (decode-norm-grp0-00-20)
  " jr   " (put-str) (opcode) 3 rshift 3 and (put-cc)
  (put-comma) (fetch-byte) c>s
  pc + (put-n16) ;

|: (decode-norm-grp0-00-10)
  (opcode) 2/ 4 and " djnzjr  " (put-str-4) (put-space)
  (fetch-byte) c>s
  pc + (put-n16) ;

|: (decode-norm-grp0-00)
  (opcode) 1 and ?exit< (decode-norm-grp0-00-01) >?
  (opcode) $20 and ?exit< (decode-norm-grp0-00-20) >?
  (opcode) $10 and ?exit< (decode-norm-grp0-00-10) >?
  (opcode) $08 and ?< " ex   af, af'" || " nop" >? (put-str) ;

|: (decode-norm-grp0-02-01)
  (opcode) 2/ 4 and " inc dec " (put-str-4) (put-space) (put-r16-sp) ;

|: (decode-norm-grp0-02)
  (opcode) 1 and ?exit< (decode-norm-grp0-02-01) >?
  " ld   " (put-str)
  (opcode) $3C and <<
    $00 of?v| " (bc), a" (put-str) |?
    $08 of?v| " a, (bc)" (put-str) |?
    $10 of?v| " (de), a" (put-str) |?
    $18 of?v| " a, (de)" (put-str) |?
    $20 of?v| (put-m16) (put-comma) (put-r16-hl-ixy) |?
    $28 of?v| (put-r16-hl-ixy) (put-comma) (put-m16) |?
    $30 of?v| (put-m16) " , a" (put-str) |?
    $38 of?v| " a, " (put-str) (put-m16) |?
  else| drop >> ;

|: (decode-norm-grp0-04)
  (opcode) $01 and 2 lshift " inc dec " (put-str-4)
  (put-space) (opcode) 3 rshift (put-r8) ;

|: (decode-norm-grp0-06-shift)
  (opcode) 2/ $1C and " rlcarrcarla rra daa cpl scf ccf " drop + (put-str-4x) ;

|: (decode-norm-grp0-06-ld)
  " ld   " (put-str) (opcode) 3 rshift (put-r8) (put-comma) (fetch-byte) (put-n8) ;

|: (decode-norm-grp0-06)
  (opcode) 1 and ?exit< (decode-norm-grp0-06-shift) >?
  (decode-norm-grp0-06-ld) ;

: (decode-norm-grp0)
  (opcode) $06 and <<
    $00 of?v| (decode-norm-grp0-00) |?
    $02 of?v| (decode-norm-grp0-02) |?
    $04 of?v| (decode-norm-grp0-04) |?
    $06 of?v| (decode-norm-grp0-06) |?
  else| drop >> ;

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
: (decode-norm)
  (opcode) $C0 and <<
    $00 of?v| (decode-norm-grp0) |?
    $40 of?v| (decode-norm-grp1) |?
    $80 of?v| (put-alu-str) (opcode) (put-r8) |?  ;; alu a,r8
  else| drop (decode-norm-grp3) >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
<published-words>

;; check ix/iy prefix
|: disasm-check-ipfx
  << $DD of?v| [char] x (ixiy):! |?
     $FD of?v| [char] y (ixiy):! |?
  else| drop >> ;

|: disasm-ipfx-pre
  (fetch-byte) dup (opcode):!
  dup $DD = over $FD = or ?<
    drop " nopx" (put-str)
    ;; one byte back
    pc 1- lo-word pc:!
    r> (dptr) over - exit
  >?
  ;; check if we have disp here
  dup 3 rshift (ixydisp-table) + c@
  1 rot 7 and lshift and ?< (fetch-byte) c>s (disp):! >? ;

;; returns disassembled text (only command, no address or bytes )
: disasm-one ( -- saddr slen )
  disbuf dup >r (dptr):!
  (ixiy):!0 (disp):!0
  (fetch-byte) dup (opcode):!
  disasm-check-ipfx
  (ixiy) ?< disasm-ipfx-pre >?
  (opcode) <<
    $CB of?v| (fetch-byte) (opcode):! (decode-cb) |?
    $ED of?v| (fetch-byte) (opcode):! (decode-ed) |?
  else| drop (decode-norm) >>
  r> (dptr) over - ;

seal-module
end-module

here swap -
 dup .( Z80 disasm size: ) . .( bytes\n)
drop
