;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Abersoft fig-FORTH extensions
;; coded by Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; .TAP writer
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<private-words>


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create basic loader

create cargador 16384 allot create; ;; should be enough for everyone

0 quan cc-pos
0 quan block#

|: cc-db,  ( byte )
  cc-pos 16384 u< not?error" out or cargador buffer"
  cc-pos cargador + c!
  cc-pos:1+! ;

|: cc-str  ( addr count )
  swap << over ?^| c@++ cc-db, 1 under- |? else| 2drop >> ;

|: cc-dec  ( num )
  lo-word base @ decimal swap <#s> cc-str base ! ;

|: create-cargador
  zxa:count-blocks dup block#:! not?error" no code blocks to write"
  cc-pos:!0
  ( line number ) 0 cc-db, 10 cc-db,
  ( line length ) 0 cc-db, 0 cc-db,
  ;; CLEAR VAL "xxx"
  " \xfd\xb0\'" cc-str zxa:clr@ cc-dec [char] " cc-db,
  ;; load blocks
  0 << zxa:next-block-from dup hi-word ?v||
    ;; :LOAD "" CODE
    " :\xef\'\'\xaf" cc-str
  ^| zxa:block-end-from | >> drop
  ;; and run
  zxa:ent@ ?<
    ;; :RANDOMIZE USR VAL "xxx"
    " :\xf9\xc0\xb0\'" cc-str zxa:ent@ cc-dec [char] " cc-db, >?
  ;; end of line
  13 cc-db,
  ;; patch line len
  cc-pos 4- cargador 2+ w! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create tape file

create tap-buf 32 allot create;

|: tap-checksum  ( addr count flag-byte -- checksum )
  >r swap << over ?^| c@++ r> xor >r 1 under- |? else| 2drop r> >> ;

|: tap-mk-header  ( addr count type )
  tap-buf 32 erase
  $13 tap-buf w!
  tap-buf 3 + c!  ;; type
  0 10 clamp
  tap-buf 4+ 10 blank
  tap-buf 4+ swap cmove ;

|: tap-header-len!  ( len )
  dup tap-buf 14 + w!
  tap-buf 18 + w! ;

|: tap-header-autostart!  ( line )
  tap-buf 16 + w! ;

|: tap-code-start!  ( zx-addr )
  tap-buf 16 + w!
  $807C tap-buf 18 + w! ;

|: tap-save-header  ( fd )
  tap-buf 2+ 18 0 tap-checksum
  tap-buf 20 + c!
  tap-buf 21 rot file:write ;

|: tap-save-data  ( addr len fd )
  over 2+ tap-buf w!  ;; total block length
  $FF tap-buf 2+ c!   ;; flag byte
  >r tap-buf 3 r@ file:write  ;; length and flag byte
  2dup r@ file:write  ;; data
  $FF tap-checksum tap-buf c!
  ( checksum byte) tap-buf 1 r> file:write ;


;; basic loader
|: tap-save-cargador  ( fd )
  create-cargador
  " cargador" 0 tap-mk-header
  cc-pos tap-header-len!
  10 tap-header-autostart!
  dup tap-save-header
  cargador cc-pos rot tap-save-data ;

|: tap-save-code  ( zx-addr-start zx-addr-end fd )
  >r
\ endcr ." BLOCK: start=$" over .hex8 ."  end=$" dup .hex8 cr
  2dup u>= ?error" invalid code block (len)"
  over hi-word over hi-word or ?error" invalid code block (addrs)"
  over - ( zx-addr-start len | fd )
  ;; tape header block
  base @ >r hex
  over <# # # # # #> string:>pad  ;; start
  dup <# # # # # [char] : #hold #> string:pad+cc  ;; size
  r> base !
endcr ." TAP: block \'" string:pad-cc@ type ." \'\n"
  string:pad-cc@ 3 tap-mk-header
  dup tap-header-len!
  over tap-code-start!
  r@ tap-save-header
endcr ." TAP: saving code...\n"
  ;; tape data block
  swap zxa:mem:ram^ swap r> tap-save-data ;

|: tap-save-code-blocks  ( fd )
  >r 0 << ( zx-addr | fd )
    zxa:next-block-from dup hi-word not?^|
      dup zxa:block-end-from ( zx-bstart zx-bend | fd )
      2dup r@ tap-save-code
      nip |?
  else| rdrop drop >> ;

@: tap-save-fd  ( fd )
  dup tap-save-cargador
endcr ." TAP: saved cargador.\n"
  tap-save-code-blocks ;

@: tap-save  ( addr count )
  file:create
  dup tap-save-fd
  file:close ;
